package com.massa.mrules.demos.advanced;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Collection;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.mrules.context.compile.CompilationLevel;
import com.massa.mrules.context.compile.MStandaloneCompilationContext;
import com.massa.mrules.context.execute.MExecutionContext;
import com.massa.mrules.exception.MConfigurationException;
import com.massa.mrules.exception.MExecutionException;
import com.massa.mrules.extensions.dsl.factory.NativeLanguageFactory;
import com.massa.mrules.factory.AbstractRuleExecutionSetFactory;
import com.massa.mrules.factory.ser.SerializationFactory;
import com.massa.mrules.factory.xml.RichXMLFactory;
import com.massa.mrules.set.IMruleExecutionSet;

/**
 * Test for a well-known race problem (cf. description in rules configuration files).<br>
 * The performance difference between mrl and xml runs shows the impact of how the rules are configured on performances.
 */
@RunWith(Parameterized.class)
public class TestRace {
	private static final Log LOG = LogFactory.getLog(TestRace.class.getName());

	final CompilationLevel level;

	public TestRace(final CompilationLevel level) {
		this.level = level;
	}

	@Parameters(name = "{0}")
	public static Collection<Object[]> configs() {
		return Arrays.asList(new Object[][] {
			{CompilationLevel.STANDARD},
			{CompilationLevel.LIGHT}
		});
	}

	private static final String XML = "com/massa/mrules/demos/advanced/TestRace.xml";
	private static final String MRL = "com/massa/mrules/demos/advanced/TestRace.mrl";

	@Test
	public void raceDemoXml() throws MConfigurationException, MExecutionException, IOException {
		final RichXMLFactory factory = new RichXMLFactory(new MStandaloneCompilationContext(this.level));
		final InputStream xml = ClassLoader.getSystemResourceAsStream(TestRace.XML);
		this.common(factory, xml, false);
	}

	@Test
	public void raceDemoMrl() throws MConfigurationException, MExecutionException, IOException {
		final NativeLanguageFactory factory = new NativeLanguageFactory(new MStandaloneCompilationContext(this.level));
		final InputStream mrl = ClassLoader.getSystemResourceAsStream(TestRace.MRL);
		this.common(factory, mrl, true);
	}

	private void common(final AbstractRuleExecutionSetFactory factory, final InputStream is, final boolean mrl)
			throws MConfigurationException, MExecutionException, IOException {
		final IMruleExecutionSet eng = factory.read(is);
		final MExecutionContext<Object> ctxt = new MExecutionContext<Object>(eng);
		ctxt.execute();

		//Time check
		TestRace.LOG.info("Beginning performance test.");
		ctxt.setLoggerName(".perf"); //to avoid debug and info.
		ctxt.setOut(null); //To avoid thousands of printings.
		//Warmup
		for (int i = 0 ; i < 5000 ; i++) {
			ctxt.execute();
		}
		//Go !
		final long t1 = System.currentTimeMillis();
		for (int i = 0 ; i < 10000 ; i++) {
			ctxt.execute();
		}
		TestRace.LOG.info("Time for 10000 tests " + (mrl ? "mrl" : "xml") + " - " + this.level + " : " + (System.currentTimeMillis() - t1));

		//Read / Write tests.
		TestRace.LOG.info("Human readable version:\n" + eng.toString());

		//Other technical tests.

		Assert.assertEquals(eng, eng.clone());

		if (! mrl) {
			//Unsupported in NativeLanguageFactory
			final StringWriter w = new StringWriter();
			factory.write(w, eng);
			TestRace.LOG.info(w.toString());
			Assert.assertEquals(eng, factory.read(new StringReader(w.toString())));
		}

		final ByteArrayOutputStream o = new ByteArrayOutputStream();
		final SerializationFactory sfactory = new SerializationFactory(this.level);
		sfactory.write(o, eng);
		Assert.assertEquals(eng, sfactory.read(new ByteArrayInputStream(o.toByteArray())));
	}

	@Test
	public void raceDemoNative() throws MConfigurationException, MExecutionException, IOException {
		this.nativeImplem(true);

		TestRace.LOG.info("Beginning native performance test.");
		final long t1 = System.currentTimeMillis();
		for (int i = 0 ; i < 10000 ; i++) {
			this.nativeImplem(false);
		}
		TestRace.LOG.info("Time for 10000 tests (native) : " + (System.currentTimeMillis() - t1));
	}

	private void nativeImplem(final boolean print) {
		for (int l = 1 ; l <= 4 ; l++) {
			for (int m = 1 ; m <= 4 ; m++) {
				for (int n = 1 ; n <= 4 ; n++) {
					for (int o = 1 ; o <= 4 ; o++) {
						if (l != m && l != n && l !=o && m != n && m != o && n != o) {
							if (
									((n == 1 && m != 2) || (n != 1 && m == 2))
									&& ((n == 2 && o != 3) || (n != 2 && o == 3))
									&& ((o == 4 && l != 2) || (o != 4 && l == 2))) {
								if (print) {
									TestRace.LOG.info("L: " + l + " ; M: " + m + " ; N: " + n + " ; o: " + o);
								}
							}
						}
					}
				}
			}
		}
	}
}
